#ifndef MSAL_H
#define MSAL_H



#define HIMMELBLAU_MAJOR 0
#define HIMMELBLAU_MINOR 2
#define HIMMELBLAU_PATCH 6


#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

#define INVALID_CRED 50126

#define REQUIRES_MFA 50076

#define INVALID_USER 50034

#define NO_CONSENT 65001

#define NO_GROUP_CONSENT 65002

#define NO_SECRET 7000218

#define AUTH_PENDING 70016

#define DEVICE_AUTH_FAIL 50155

typedef enum MSAL_ERROR {
  INVALID_JSON,
  INVALID_BASE64,
  INVALID_REGEX,
  INVALID_PARSE,
  ACQUIRE_TOKEN_FAILED,
  GENERAL_FAILURE,
  REQUEST_FAILED,
  AUTH_TYPE_UNSUPPORTED,
  TPM_FAIL,
  URL_FORMAT_FAILED,
  DEVICE_ENROLLMENT_FAIL,
  CRYPTO_FAIL,
  NOT_IMPLEMENTED,
  CONFIG_ERROR,
  MFA_POLL_CONTINUE,
  SUCCESS,
  INVALID_POINTER,
  NO_MEMORY,
} MSAL_ERROR;

typedef enum TracingLevel {
  ERROR,
  WARN,
  INFO,
  DEBUG,
  TRACE,
} TracingLevel;

typedef struct BoxedDynTpm BoxedDynTpm;

typedef struct BrokerClientApplication BrokerClientApplication;

typedef struct DeviceAuthorizationResponse DeviceAuthorizationResponse;

typedef struct EnrollAttrs EnrollAttrs;

typedef struct LoadableIdentityKey LoadableIdentityKey;

typedef struct LoadableMachineKey LoadableMachineKey;

typedef struct LoadableMsOapxbcRsaKey LoadableMsOapxbcRsaKey;

typedef struct MFAAuthContinue MFAAuthContinue;

typedef struct MachineKey MachineKey;

typedef struct SealedData SealedData;

typedef struct UserToken UserToken;

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void raw_serialized_free(uint8_t *input, uintptr_t len);

enum MSAL_ERROR set_global_tracing_level(enum TracingLevel level);

/**
 * Initialize a TPM struct
 *
 * # Arguments
 *
 * * `tcti_name` - An optional TPM Transmission Interface. If this paramenter
 *   is NULL, a Soft Tpm is initialized.
 *
 * * `out` - An output parameter which will contain the initialized
 *   BoxedDynTpm.
 *
 * # Safety
 *
 * The calling function must ensure that the `tcti_name` is either a valid c
 * string, or NULL. It must also ensure that `out` is a valid c BoxedDynTpm
 * double pointer.
 */
enum MSAL_ERROR tpm_init(const char *tcti_name, struct BoxedDynTpm **out);

/**
 * # Safety
 *
 * The calling function must ensure that `out` is a valid c
 * char double pointer.
 */
enum MSAL_ERROR auth_value_generate(char **out);

/**
 * # Safety
 *
 * The calling function must ensure that the `tpm` is a valid BoxedDynTpm,
 * that `auth_value` is a valid C string, and that `out` is a valid c
 * LoadableMachineKey double pointer.
 */
enum MSAL_ERROR tpm_machine_key_create(struct BoxedDynTpm *tpm,
                                       const char *auth_value,
                                       struct LoadableMachineKey **out);

/**
 * # Safety
 *
 * The calling function must ensure that the `tpm` is a valid BoxedDynTpm,
 * that `auth_value` is a valid C string, that `exported_key` is a valid
 * LoadableMachineKey, and that `out` is a valid C MachineKey double pointer.
 */
enum MSAL_ERROR tpm_machine_key_load(struct BoxedDynTpm *tpm,
                                     const char *auth_value,
                                     struct LoadableMachineKey *exported_key,
                                     struct MachineKey **out);

/**
 * Create an instance of an application.
 *
 * # Arguments
 *
 * * `authority` - A URL that identifies a token authority. It should
 *   be of the format <https://login.microsoftonline.com/your_tenant> By
 *   default, we will use <https://login.microsoftonline.com/common>.
 *
 * * `transport_key` - An optional LoadableMsOapxbcRsaKey transport key
 *   from enrolling the device.
 *
 * * `cert_key` - An optional LoadableIdentityKey which was used to create
 *   the enrollment CSR.
 *
 * * `out` - An output parameter which will contain the initialized
 *   BrokerClientApplication.
 *
 * NOTE: If `transport_key` and `cert_key` are not provided from a previous
 * device enrollment, then enrollment will be required.
 *
 * # Safety
 *
 * The calling function must ensure that the `authority` is either a valid c
 * string, or NULL. It must also ensure that `transport_key` is a valid c
 * LoadableMsOapxbcRsaKey pointer or NULL, that `cert_key` is a valid c
 * LoadableIdentityKey pointer or NULL, and that `out` is a valid c
 * BrokerClientApplication double pointer.
 */
enum MSAL_ERROR broker_init(const char *authority,
                            struct LoadableMsOapxbcRsaKey *transport_key,
                            struct LoadableIdentityKey *cert_key,
                            struct BrokerClientApplication **out);

/**
 * Initialize attributes for device enrollment
 *
 * # Arguments
 *
 * * `target_domain` - The domain to be enrolled in.
 *
 * * `device_display_name` - An optional chosen display name for the
 *   enrolled device. Defaults to the system hostname.
 *
 * * `device_type` - An optional device type. Defaults to 'Linux'. This
 *   effects which Intune policies are distributed to the client.
 *
 * * `join_type` - A join type. Possible values are:
 *     - 0: Azure AD join.
 *     - 4: Azure AD register only.
 *     - 6: Azure AD hybrid join.
 *     - 8: Azure AD join.
 *
 * * `os_version` - An optional OS version. Defaults to the contents of
 *   /etc/os-release.
 *
 * * `out` - A new EnrollAttrs for device enrollment.
 *
 * # Safety
 *
 * The calling function must ensure that `target_domain` is a valid c string,
 * that `device_display_name` is either a valid c string or NULL, that
 * `device_type` is either a valid c string or NULL, that `os_version` is
 * either a valid c string or NULL, and that `out` is a valid c EnrollAttrs
 * double pointer.
 */
enum MSAL_ERROR enroll_attrs_init(const char *target_domain,
                                  char *device_display_name,
                                  char *device_type,
                                  int join_type,
                                  char *os_version,
                                  struct EnrollAttrs **out);

/**
 * Enroll the device in the directory.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `token` - Token obtained via either
 *   acquire_token_by_username_password_for_device_enrollment
 *   or acquire_token_by_device_flow.
 *
 * * `attrs` - A EnrollAttrs created by a call to `enroll_attrs_init`. This
 *   parameter is consumed and destroyed by the recipient.
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `out_transport_key` - A LoadableMsOapxbcRsaKey transport key output
 *   parameter.
 *
 * * `out_cert_key` - A LoadableIdentityKey certificate key output parameter.
 *
 * * `out_device_id` - An output parameter which is the device id of the
 *   enrolled device.
 *
 * # Safety
 *
 * The calling function should ensure that `client`, `token`, `attrs`, `tpm`,
 * and `machine_key` are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_enroll_device(struct BrokerClientApplication *client,
                                     struct UserToken *token,
                                     struct EnrollAttrs *attrs,
                                     struct BoxedDynTpm *tpm,
                                     struct MachineKey *machine_key,
                                     struct LoadableMsOapxbcRsaKey **out_transport_key,
                                     struct LoadableIdentityKey **out_cert_key,
                                     char **out_device_id);

/**
 * Gets a token for a given resource via user credentials.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `username` - Typically a UPN in the form of an email address.
 *
 * * `password` - The password.
 *
 * * `scopes` - An array of scopes requested to access a protected API (a
 *   resource).
 *
 * * `request_resource` - A resource for obtaining an access token.
 *   Default is the MS Graph API (00000002-0000-0000-c000-000000000000).
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `out` - A UserToken containing an access_token.
 *
 * # Safety
 *
 * The calling function should ensure that `client`, `username`, `password`,
 * `scopes`, `tpm`, and `machine_key` are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_acquire_token_by_username_password(struct BrokerClientApplication *client,
                                                          const char *username,
                                                          const char *password,
                                                          const char *const *scopes,
                                                          int scopes_len,
                                                          const char *request_resource,
                                                          struct BoxedDynTpm *tpm,
                                                          struct MachineKey *machine_key,
                                                          struct UserToken **out);

/**
 * Acquire token(s) based on a refresh token (RT) obtained from elsewhere.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `refresh_token` - The old refresh token, as a string.
 *
 * * `scopes` - The scopes associated with this old RT.
 *
 * * `request_resource` - A resource for obtaining an access token.
 *   Default is the MS Graph API (00000002-0000-0000-c000-000000000000).
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `out` - A UserToken, which means migration was successful.
 *
 * # Safety
 *
 * The calling function must ensure that `client`, `tpm`, `machine_key`, are
 * valid pointers to their respective types.
 */
enum MSAL_ERROR broker_acquire_token_by_refresh_token(struct BrokerClientApplication *client,
                                                      const char *refresh_token,
                                                      const char *const *scopes,
                                                      int scopes_len,
                                                      const char *request_resource,
                                                      struct BoxedDynTpm *tpm,
                                                      struct MachineKey *machine_key,
                                                      struct UserToken **out);

/**
 * Gets a token for enrollment via user credentials.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `username` - Typically a UPN in the form of an email address.
 *
 * * `password` - The password.
 *
 * * `out` - A UserToken containing an access_token.
 *
 * # Safety
 *
 * The calling function must ensure that `client` is a valid pointer.
 */
enum MSAL_ERROR broker_acquire_token_by_username_password_for_device_enrollment(struct BrokerClientApplication *client,
                                                                                const char *username,
                                                                                const char *password,
                                                                                struct UserToken **out);

/**
 * Initiate a Device Flow instance for enrollment, which will be
 * used in acquire_token_by_device_flow.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `out` - A DeviceAuthorizationResponse containing a user_code key,
 *   among others
 *
 * # Safety
 *
 * The calling function must ensure that `client` is a valid pointer.
 */
enum MSAL_ERROR broker_initiate_device_flow_for_device_enrollment(struct BrokerClientApplication *client,
                                                                  struct DeviceAuthorizationResponse **out);

/**
 * Obtain token for enrollment by a device flow object, with customizable
 * polling effect.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `flow` - A DeviceAuthorizationResponse previously generated by
 * initiate_device_flow.
 *
 * * `out` - A UserToken containing an access_token.
 *
 * # Safety
 *
 * The calling function must ensure that `client` is a valid pointer.
 */
enum MSAL_ERROR broker_acquire_token_by_device_flow(struct BrokerClientApplication *client,
                                                    struct DeviceAuthorizationResponse *flow,
                                                    struct UserToken **out);

/**
 * Check if a user exists in Azure Entra ID
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `username` - Typically a UPN in the form of an email address.
 *
 * * `out` - true|false
 *
 * # Safety
 *
 * The calling function must ensure that `client` is a valid pointer.
 */
enum MSAL_ERROR broker_check_user_exists(struct BrokerClientApplication *client,
                                         const char *username,
                                         bool *out);

/**
 * Initiate an MFA flow for enrollment via user credentials.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `username` - Typically a UPN in the form of an email address.
 *
 * * `password` - The password.
 *
 * * `out` - A MFAAuthContinue containing the information needed to continue the
 *   authentication flow.
 *
 * # Safety
 *
 * The calling function should ensure that `client`, `username`, and
 * `password`, are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_initiate_acquire_token_by_mfa_flow_for_device_enrollment(struct BrokerClientApplication *client,
                                                                                const char *username,
                                                                                const char *password,
                                                                                struct MFAAuthContinue **out);

/**
 * Obtain token by a MFA flow object.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `username` - Typically a UPN in the form of an email address.
 *
 * * `auth_data` - An optional token received for the MFA flow (some MFA
 *   flows do not require input). If this MFA type does not require input,
 *   this MUST be NULL.
 *
 * * `poll_attempt` - The polling attempt number. If this MFA type requires
 *   input, this should be 0.
 *
 * * `flow` - A MFAAuthContinue previously generated by
 * initiate_acquire_token_by_mfa_flow.
 *
 * * `out` - A UserToken containing an access_token.
 * # Safety
 *
 * The calling function should ensure that `client`, `username`, `auth_data`,
 * `poll_attempt`, and `flow` are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_acquire_token_by_mfa_flow(struct BrokerClientApplication *client,
                                                 const char *username,
                                                 const char *auth_data,
                                                 int poll_attempt,
                                                 struct MFAAuthContinue *flow,
                                                 struct UserToken **out);

/**
 * Get the msg from a MFAAuthContinue flow
 *
 * # Safety
 *
 * The calling function should ensure that `flow` is a valid MFAAuthContinue
 * pointer, and that `out` is a valid double c_char pointer.
 */
enum MSAL_ERROR mfa_auth_continue_msg(struct MFAAuthContinue *flow, char **out);

/**
 * Get the mfa_method from a MFAAuthContinue flow
 *
 * # Safety
 *
 * The calling function should ensure that `flow` is a valid MFAAuthContinue
 * pointer, and that `out` is a valid double c_char pointer.
 */
enum MSAL_ERROR mfa_auth_continue_mfa_method(struct MFAAuthContinue *flow, char **out);

/**
 * Get the polling_interval from a MFAAuthContinue flow
 *
 * If no polling_interval was defined by Entra ID, this function returns -1.
 *
 * # Safety
 *
 * The calling function should ensure that `flow` is a valid MFAAuthContinue
 * pointer.
 */
int mfa_auth_continue_polling_interval(struct MFAAuthContinue *flow);

/**
 * Get the max_poll_attempts from a MFAAuthContinue flow
 *
 * If no max_poll_attempts was defined by Entra ID, this function returns -1.
 *
 * # Safety
 *
 * The calling function should ensure that `flow` is a valid MFAAuthContinue
 * pointer.
 */
int mfa_auth_continue_max_poll_attempts(struct MFAAuthContinue *flow);

/**
 * Gets a Primary Refresh Token (PRT) via user credentials.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `username` - Typically a UPN in the form of an email address.
 *
 * * `password` - The password.
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `out` - An encrypted PrimaryRefreshToken, containing a refresh_token and tgt.
 *
 * # Safety
 *
 * The calling function should ensure that `client`, `username`, `password`,
 * `tpm`, and `machine_key` are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_acquire_user_prt_by_username_password(struct BrokerClientApplication *client,
                                                             const char *username,
                                                             const char *password,
                                                             struct BoxedDynTpm *tpm,
                                                             struct MachineKey *machine_key,
                                                             struct SealedData **out);

/**
 * Gets a Primary Refresh Token (PRT) via a refresh token (RT) obtained
 * previously.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `refresh_token` - The old refresh token, as a string.
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `out` - An encrypted PrimaryRefreshToken, containing a refresh_token and tgt.
 * # Safety
 *
 * The calling function should ensure that `client`, `refresh_token`, `tpm`,
 * and `machine_key` are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_acquire_user_prt_by_refresh_token(struct BrokerClientApplication *client,
                                                         const char *refresh_token,
                                                         struct BoxedDynTpm *tpm,
                                                         struct MachineKey *machine_key,
                                                         struct SealedData **out);

/**
 * Given the primary refresh token, this method requests an access token.
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `sealed_prt` -  An encrypted primary refresh token that was
 *   previously received from the server.
 *
 * * `scopes` - The scopes that the client requests for the access token.
 *
 * * `request_resource` - A resource for obtaining an access token.
 *   Default is the MS Graph API (00000002-0000-0000-c000-000000000000).
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `out` - A UserToken containing an access_token.
 *
 * # Safety
 *
 * The calling function should ensure that `client`, `sealed_prt`, `scope`,
 * `request_resource`, `tpm`, and `machine_key` are valid pointers to their
 * respective types.
 */
enum MSAL_ERROR broker_exchange_prt_for_access_token(struct BrokerClientApplication *client,
                                                     struct SealedData *sealed_prt,
                                                     const char *const *scopes,
                                                     int scopes_len,
                                                     const char *request_resource,
                                                     struct BoxedDynTpm *tpm,
                                                     struct MachineKey *machine_key,
                                                     struct UserToken **out);

/**
 * Given the primary refresh token, this method requests a new primary
 * refresh token
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `sealed_prt` -  An encrypted primary refresh token that was
 *   previously received from the server.
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `request_tgt` - Whether to include a request for a TGT.
 *
 * * `out` - An encrypted PrimaryRefreshToken, containing a refresh_token
 *   and optionally a tgt. The session key is copied from the old PRT.
 *
 * # Safety
 *
 * The calling function should ensure that `client`, `sealed_prt`,
 * `tpm`, and `machine_key` are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_exchange_prt_for_prt(struct BrokerClientApplication *client,
                                            struct SealedData *sealed_prt,
                                            struct BoxedDynTpm *tpm,
                                            struct MachineKey *machine_key,
                                            int request_tgt,
                                            struct SealedData **out);

/**
 * Provision a new Hello for Business Key
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `token` - Token obtained via either
 *   acquire_token_by_username_password_for_device_enrollment
 *   or acquire_token_by_device_flow.
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `pin` - The PIN code which will be used to unlock the key.
 *
 * * `out` - Either the existing LoadableIdentityKey, or a new created
 *   key if none was provided.
 *
 * # Safety
 *
 * The calling function should ensure that `client`, `token`, `tpm`,
 * `machine_key`, and `pin` are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_provision_hello_for_business_key(struct BrokerClientApplication *client,
                                                        struct UserToken *token,
                                                        struct BoxedDynTpm *tpm,
                                                        struct MachineKey *machine_key,
                                                        const char *pin,
                                                        struct LoadableIdentityKey **out);

/**
 * Gets a token for a given resource via a Hello for Business Key
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `username` - Typically a UPN in the form of an email address.
 *
 * * `key` - A LoadableIdentityKey provisioned using
 *   provision_hello_for_business_key.
 *
 * * `scopes` - Scopes requested to access a protected API (a resource).
 *
 * * `request_resource` - A resource for obtaining an access token.
 *   Default is the MS Graph API (00000002-0000-0000-c000-000000000000).
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `pin` - The PIN code required to unlock the key.
 *
 * * `out` - A UserToken containing an access_token.
 *
 * # Safety
 *
 * The calling function should ensure that `client`, `username`, `key`, `scopes`,
 * `request_resource`, `tpm`, `machine_key`, and `pin` are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_acquire_token_by_hello_for_business_key(struct BrokerClientApplication *client,
                                                               const char *username,
                                                               struct LoadableIdentityKey *key,
                                                               const char *const *scopes,
                                                               int scopes_len,
                                                               const char *request_resource,
                                                               struct BoxedDynTpm *tpm,
                                                               struct MachineKey *machine_key,
                                                               const char *pin,
                                                               struct UserToken **out);

/**
 * Gets a Primary Refresh Token (PRT) via a Hello for Business Key
 *
 * # Arguments
 *
 * * `client` - A BrokerClientApplication created by a call to
 *   `broker_init`.
 *
 * * `username` - Typically a UPN in the form of an email address.
 *
 * * `key` - A LoadableIdentityKey provisioned using
 *   provision_hello_for_business_key.
 *
 * * `tpm` - The tpm object.
 *
 * * `machine_key` - The TPM MachineKey associated with this application.
 *
 * * `pin` - The PIN code required to unlock the key.
 *
 * * `out` - An encrypted PrimaryRefreshToken, containing a refresh_token and tgt.
 *
 * # Safety
 *
 * The calling function should ensure that `client`, `username`, `key`, `tpm`,
 * `machine_key`, and `pin` are valid pointers to their respective types.
 */
enum MSAL_ERROR broker_acquire_user_prt_by_hello_for_business_key(struct BrokerClientApplication *client,
                                                                  const char *username,
                                                                  struct LoadableIdentityKey *key,
                                                                  struct BoxedDynTpm *tpm,
                                                                  struct MachineKey *machine_key,
                                                                  const char *pin,
                                                                  struct SealedData **out);

/**
 * # Safety
 *
 * The calling function must ensure that the `token` raw pointer is valid and
 * can be dereferenced, and that `out` is a valid pointer to a char*.
 */
enum MSAL_ERROR user_token_refresh_token(struct UserToken *token, char **out);

/**
 * # Safety
 *
 * The calling function must ensure that the `token` raw pointer is valid and
 * can be dereferenced, and that `out` is a valid pointer to a char*.
 */
enum MSAL_ERROR user_token_access_token(struct UserToken *token, char **out);

/**
 * # Safety
 *
 * The calling function must ensure that the `token` raw pointer is valid and
 * can be dereferenced, and that `out` is a valid pointer to a char*.
 */
enum MSAL_ERROR user_token_tenant_id(struct UserToken *token, char **out);

/**
 * # Safety
 *
 * The calling function must ensure that the `token` raw pointer is valid and
 * can be dereferenced, and that `out` is a valid pointer to a char*.
 */
enum MSAL_ERROR user_token_spn(struct UserToken *token, char **out);

/**
 * # Safety
 *
 * The calling function must ensure that the `token` raw pointer is valid and
 * can be dereferenced, and that `out` is a valid pointer to a char*.
 */
enum MSAL_ERROR user_token_uuid(struct UserToken *token, char **out);

/**
 * # Safety
 *
 * The calling function must ensure that the `token` raw pointer is valid and
 * can be dereferenced, and that `out` is a valid pointer to a bool.
 */
enum MSAL_ERROR user_token_amr_mfa(struct UserToken *token, bool *out);

/**
 * # Safety
 *
 * The calling function must ensure that the `client` raw pointer is valid and
 * can be dereferenced.
 */
void broker_free(struct BrokerClientApplication *client);

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void string_free(char *input);

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void machine_key_free(struct MachineKey *input);

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void loadable_machine_key_free(struct LoadableMachineKey *input);

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void loadable_ms_oapxbc_rsa_key_free(struct LoadableMsOapxbcRsaKey *input);

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void loadable_identity_key_free(struct LoadableIdentityKey *input);

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void user_token_free(struct UserToken *input);

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void tpm_free(struct BoxedDynTpm *input);

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void mfa_auth_continue_free(struct MFAAuthContinue *input);

/**
 * # Safety
 *
 * The calling function must ensure that the `input` raw pointer is valid and
 * can be dereferenced.
 */
void sealed_data_free(struct SealedData *input);

enum MSAL_ERROR serialize_loadable_identity_key(struct LoadableIdentityKey *value,
                                                char **out_buf, uintptr_t *out_len);
enum MSAL_ERROR deserialize_loadable_identity_key(uint8_t *in_buf, uintptr_t in_len,
                                                  struct LoadableIdentityKey **out);
enum MSAL_ERROR serialize_loadable_machine_key(struct LoadableMachineKey *value,
                                               char **out_buf, uintptr_t *out_len);
enum MSAL_ERROR deserialize_loadable_machine_key(uint8_t *in_buf, uintptr_t in_len,
                                                 struct LoadableMachineKey **out);
enum MSAL_ERROR serialize_loadable_ms_oapxbc_rsa_key(struct LoadableMsOapxbcRsaKey *value,
                                                     char **out_buf, uintptr_t *out_len);
enum MSAL_ERROR deserialize_loadable_ms_oapxbc_rsa_key(uint8_t *in_buf, uintptr_t in_len,
                                                       struct LoadableMsOapxbcRsaKey **out);
enum MSAL_ERROR serialize_sealed_data(struct SealedData *value,
                                      char **out_buf, uintptr_t *out_len);
enum MSAL_ERROR deserialize_sealed_data(uint8_t *in_buf, uintptr_t in_len,
                                        struct SealedData **out);

#endif /* MSAL_H */
